/*VagueDenoiser plugin for Avisynth -- a wavelet denoiser plugin
  VagueDenoiser (C) 2003, LeFunGus
  VagueDenoiser assembly parts, debug and adaptative thresholding (C) Kurosu
  WienerChop mode, 9/7 and 6/10 wavelets, AutoThreshold, aux clip (c)2004-2005 Fizick 

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  The author can be contacted at:
  lefungus@altern.org

  Thanks to Geoff Davis for his wavelet Transform Coder Construction Kit
  Look at http://www.geoffdavis.net/ for the full package.
  Thanks to everyone at Doom9.org for their help.

  V0.12: 1st release
  -Precision problems corrected, 0 for threshold is now lossless.

  V0.13:
  -Hard thresholding method enabled.
  -Cosmetic changes.
  -Avisynth parameters changed ( VagueDenoiser(threshold=0.8,wavelet=[0,1],method=[0,1].
   
  V0.2:
  -Implemented many new wavelet filters, thanks to the wavelet transform coder construction kit.
   look at http://www.geoffdavis.net/ for more informations.
  -Implemented nsteps parameter, that allow you to use n steps in the selected wavelet tranform.
   Some nsteps values could produce unvalid results, reducing this value generally solve the problem.
  -Little optimizations from Shodan and Bidoche.

  V0.22:
  -Implemented YUY2 colorspace.
  -Optionnal chroma filtering, (chroma=true/false).
  -Cosmetic changes, code cleaned.
  -Html documentation.

  V0.23:
  -Code cleaned, filters class removed.
  -Better parameters for compilation. Should really works on every cpu now.
  -Removed all filters except Brislawn 10/10, so filter parameter has been removed.(it's like filter=7).
  -A little speed increase (3-5fps on a 640x256 frame with nsteps=6).

  V0.24:
  -Another speed increase, some critical loops unrolled (To infinity and beyond!)
  
  V0.241:
  -height and width must be mod4, added errors messages if not
  -Cleaned include thanks to Kurosu

  V0.242:
  -Removed restrictions on width/height, fixed bugs
  -Some improvements from Kurosu

  V0.243:
  -Code cleaned

  V0.25:
  -Little speed increase (due to some little profiles and vectorizations)
  -added defaults
  -added a visual.net compiled dll for compatibility purposes. This dll is slower

  V0.26 (Kurosu)
  - Merged (C++ frenzy) all assembly parts from Kurosu's version into one dll
  - Hence, great speed increase
  - All improvements are available for YV12 only

  V0.26.1 (Kurosu)
  - Merge from Sh0dan on copy
  - Cleaner and safer management from Bidoche
  - Fixed a crash that may have affected P4 users (Athlons with SSE support weren't
    affected as 3DNow! code, being the fatest, is always selected).
  - Added debug output and NOOP threshold method (for debug purpose)

  V0.27.0 (Kurosu)
  - Continued integration and added framework for Haar wavelet (still not functionnal,
	deactivated)
  - All optimizations proposed by Sh0dan, ARDA and Bidoche
  - Some registers reuse (AMD-64 gonna rock) for some more speed (5%)
  - Added Qian thresholding
    
  V0.27.1 (Kurosu)
  - Fixed YUY2 mode
  - Added RGB24 and RGB32 mode
  - The 3 above modes are unlikely to get any speed optimization
    
  V0.28.0 (Kurosu)
  - Fixed all modes to properly process what needs to be processed (small speedup)
  - Unrolled float2byte conversion, little speed-up
  - Added interlaced (see option with that name) support to YV12 colorspace
  - Workspace compatible with ICL7. dll isn't compiled in that mode for
	legal reasons (I don't own ICL)
  - Decteted a major slowdown for MOD64 width. Avoid them, as applying this filter
	before resizing might be faster in fact (particularly true for widths of 512)
    
  V0.28.1 (Kurosu)
  - MOD64 width slowdown is due to something needing a dreadfull rewrite, so
	no fix
  - Implemented cleaner support of YUY2 and RGB24/32 
  - Interlaced mode now works in all mode (processing in interlaced mode is a little
	bit faster but you loose precision)
  - Certainly the last of its branch

  V0.28.2 (Kurosu) 
  Ported to nasm the assembly parts 

  V0.29 (Kurosu) 
  More nasm 
  More bug fixes 

  V0.30 (Fizick)
  Added (restored) Cohen-Daubechies-Feauveau 9/7 wavelet (C version only) and wavelet selection option
  Added WienerChop two-pass denoising mode using two different wavelet basises.
  Replaced boolean "chroma" parameter to float "chromaT" as threshold value for chroma planes
  Added automatic (adaptive) threshold estimation
  3DNow mode of hard thesholding temporary replaced by SSE or C versions.
  Fixed possible memory leakage bug for non YV12 modes

  V.0.31 (Fizick) 
  Added noise ratio parameter for Wiener pass
  3DNow mode of Qian thesholding temporary replaced by SSE or C versions.
  Fixed interlaced mode. Seems it now works.

  V.0.32 (Released by Fizick, but part of work was done by Kurosu)
  Added Villasenor-Belzer-Liao 6/10 wavelet (not optimized C version only)
  Changed estimation wavelets to more optimal pairs for WienerChop mode.
  3DNow optimized version of WienerChop and AutoThreshold - thanks to Kurosu.
  3DNow mode of hard and Qian thesholding re-enabled after some bugs were fixed by Kurosu and Fizick.
  Small speed increasing mainly due to copy reverse order 
  (but big slowdown for mod64 still exist).
  SSE float-byte conversion has bug for some width, temporary replaced to C version.
  Ported to NASM memcopy assembly function (no more non-NASM assembly).

  V.0.32.1 - July 10, 2004 (Fizick)
  Fixed bug in copy function, introduced in v.0.32
  Re-enabled SSE optimized float-byte and byte-float conversion after fixing some bugs, 

  V.0.33 - July 11, 2004 (Fizick)
  Fixed slowdown for mod64 width by padding (thanks to MfA)
  Added partial denoising mode (by blending with source)
  Fixed bug with AutoThreshold for 3DNow.
  AutoThreshold is now also dependent from "wratio" parameter.
  Added messages for Debugview utility
  Set some parameters default values to more optimal (for me?):
     thresh=0 (auto), method=3, nsteps=4, wavelet=1, percent=75.

  V.0.33.1 - July 13, 2004 (Fizick)
  Fixed bug with mod64 width for SSE (by pad increasing from 2 to 4)
  Small change of mem_set for debug

  V.0.33.2 - July 17, 2004 (Fizick)
  Fixed bug with default values (thanks to SILICON for report).

  V.0.33.3 - July 21, 2004 (Fizick)
  Fixed bug with YUY2 (introduced in v.0.32.1).

  V.0.33.4 - August 23, 2004 (Fizick)
  Fixed bug with mirrorred padded pixels (thanks to Eugen65 for report).

  V.0.33.5 - September 28, 2004 (Fizick) 
  Fixed bug with AutoThreshold for Interlaced Wiener mode (thanks to Viperzahn and LigH for report).

  V.0.33.6 - October 13, 2004 (Fizick) 
  Nsteps parameter now is auto-limited to max admissible value if input too big or =0 (don't worry, Viperzahn :)
  Some improving of exception handling (try-catch-throw).

  V.0.33.7 - October 16, 2004 (Fizick) 
  Decreased max admissible value of nsteps to fix some internal bug (or feature)
  (thanks to Viperzahn for insistent report :)

  V.0.34 - November 24, 2004 (Fizick)
  Added auxiliary (some prefiltered) clip for first pass of Wiener mode.

  V.0.34.1 - December 19, 2004 (Fizick)
  Some fix for mem_set compatibility with old CPU (P2).

  V.0.34.2 - March 10, 2005 (Fizick)
  Fixed small bug (blue dot) for chromaT>=0 in YV12 for Athlon (thanks to Pavico for report)

  v.0.34.3 - Septerber 11, 2005 
  Fixed authothreshold algo

  V.0.35 - September 17, 2005 (Fizick)
  Improving soft method=1 - do not filter lowest level anymore, get more stable mean picture intencity.
  Re-enabled method=2 of multilevel subband adaptive thresholding, implemented as NormalShrink method;
  Added method=4 uniformly smooth shrinking function;
  Replaced partial denoising blend mode by shrinking of small coefficients for all thresholding functions; 
  Changed default percent=85;
  Implemented new SSE versions for method=0,3,4 and C versions for rest (3DNow disabled with 4% speed decreasing); 
  Added 3DNow optimized versions for transform of wavelet 1 and 3;

  v.0.35.1 - september 26,2005
  Added YUY2 and RGB format for auxclip

*/

#include "VagueDenoiser.h"

//#define OP(a) logf(1+a)

VagueDenoiser::VagueDenoiser(PClip _child1, double _lumaT, int _method, int _nsteps,
							 double _chromaT, bool _dbg, bool _itlc,
							 int _wlt,  bool _Wiener, float _wratio, float _percent, PClip _auxClip,
							 IScriptEnvironment* env) :
GenericVideoFilter(_child1), method(_method), nsteps(_nsteps), 
	debug(_dbg), itlc(_itlc), wlt(_wlt),
	Wiener(_Wiener), wratio(_wratio), percent(_percent), auxClip(_auxClip), VDEnv(env) 
{
	lumaT = (float)_lumaT;
	chromaT = (float)_chromaT;
//	if (thresh<0.0f)
//		env->ThrowError("VagueDenoiser: positive thresholds only"); // Removed in 0.31.
	if (_method<-1 || _method>4)
		env->ThrowError("VagueDenoiser: Invalid threshold method");

	if (percent<0 || percent>100)
		env->ThrowError("VagueDenoiser: Percent of denoising must be from 0 to 100");

	width=vi.width;
	height=vi.height;

	if (vi.IsYV12())
	{
		heightU=height>>1; //   1/2
		widthU=width>>1; //    1/2
	}
	else
	{
		if (vi.IsYUY2())
		{
			widthU=width>>1;//    1/2
			heightU=height;
		}
		else if (vi.IsRGB())
		{
			widthU=width;
			heightU=height;
		}
		else env->ThrowError("VagueDenoiser: YV12, YUY2, RGB24 and RGB32 (so what are you using?) only");

	}

	if (itlc)
	{
		if (heightU%2)
			env->ThrowError("VagueDenoiser: interlaced mode requires MOD2 height for all planes");
	}

	int widthPadded = (width%64) ? width : width + WIDTHPAD; // v0.35
	int widthUPadded = (widthU%64) ? widthU : widthU + WIDTHPAD;// v0.35

	pitch = ((width+7)/8)*8; // added in v.0.34.2 
	bufSize = pitch*height;// changed in v.0.34.2

	pitchU = ((width+7)/8)*8; // added in v.0.34.2 
	bufSizeU = pitchU*heightU;// changed v.0.34.2

	// more correct nsteps checking code in v.0.33.6

	int nstepsmax; // caclulate max possible nsteps
	for (nstepsmax = 1; nstepsmax < 15; nstepsmax++)
	{
		if (itlc) // interlaced
		{
			if (chromaT < 0) 
			{ // luma only, no chroma
				if (pow(2,nstepsmax+1)>=height || pow(2,nstepsmax)>=widthPadded) // changed to >= in 0.33.7, v0.35
					break;
			}
			else
			{ // chroma too
				if (pow(2,nstepsmax+1)>=heightU || pow(2,nstepsmax)>=widthUPadded)// changed to >= in 0.33.7, v0.35
					break;
			}
		}
		else // progressive
		{ 
			if (chromaT < 0) 
			{ // luma only, no chroma
				if (pow(2,nstepsmax)>=height || pow(2,nstepsmax)>=widthPadded) // changed to >= in 0.33.7, v0.35
					break;
			}
			else
			{ // chroma too
				if (pow(2,nstepsmax)>=heightU || pow(2,nstepsmax)>=widthUPadded) // changed to >= in 0.33.7, v0.35
					break;
			}
		}
	}
	nstepsmax--; // correct by 1 to get true value

	nstepsmax--; // decrease maximum admissible value to fix some internal bug (feature) - added in v 0.33.7

	if (nstepsmax < 1 )
		env->ThrowError("VagueDenoiser: clip dimensions too low!"); // added in 0.33.7
	
	if (nsteps > nstepsmax || nsteps < 1)
	{
		nsteps = nstepsmax; // limit nsteps by maximum admissible value - added in v.0.33.6
//		env->ThrowError("VagueDenoiser: clip dimensions too low to permit %i steps", _nsteps); // disabled in v.0.33.6
	}


//	if (debug)
//		env->ThrowError("VagueDenoiser: sorry, debug mode disable in this release"); // Enabled in 0.32. Let's people see!
	
	if (!vi.IsYV12())
	{
		buffY= (unsigned char *) _mm_malloc(bufSize*sizeof(unsigned char),128); // replaced to bufSize in 0.32
		buffU= (unsigned char *) _mm_malloc(bufSizeU*sizeof(unsigned char),128);
		buffV= (unsigned char *) _mm_malloc(bufSizeU*sizeof(unsigned char),128);
	}
		
	block= (float*) _mm_malloc((bufSize + height*16)*sizeof(float),128); // replaced to bufSize in 0.32, increased to enable width padding in v.0.33 

	if (Wiener) block1= (float*) _mm_malloc((bufSize + height*16)*sizeof(float),128); // v0.30 by Fizick

	/* Buffer for interlaced conversion and whatnot */
	if (itlc)
	{
		buffer = (unsigned char *) _mm_malloc(bufSize*sizeof(char),32);
	}

	child->SetCacheHints(CACHE_RANGE,0); // add in v.0.30

	useAuxClip = false; // aux clip added in v.0.34
	if (auxClip && Wiener==true)
	{
//		if (!vi.IsYV12()) env->ThrowError("VagueDenoiser: aux clip is supported for YV12 only!"); // disabled in v.035.1
		VideoInfo auxvi = auxClip->GetVideoInfo();
		if (vi.IsSameColorspace(auxvi) && vi.height == auxvi.height && vi.width == auxvi.width && vi.num_frames == auxvi.num_frames)
		{
			useAuxClip = true;
			auxClip->SetCacheHints(CACHE_RANGE,0);
			if (!vi.IsYV12())
			{
				auxbuffY= (unsigned char *) _mm_malloc(bufSize*sizeof(unsigned char),128); // replaced to bufSize in 0.32
				auxbuffU= (unsigned char *) _mm_malloc(bufSizeU*sizeof(unsigned char),128);
				auxbuffV= (unsigned char *) _mm_malloc(bufSizeU*sizeof(unsigned char),128);
			}
		}
		else
			env->ThrowError("VagueDenoiser: aux clip must have same properties!"); 
	}

	int height_itlc = (itlc) ? height/2 : height;
	int heightU_itlc = (itlc) ? heightU/2 : heightU;
	if (method == 2) // v0.35
	{
		LUTY = (unsigned char*)_mm_malloc(widthPadded*height_itlc*sizeof(char),128);
		InitTable(widthPadded, height_itlc, nsteps, LUTY);
		LUTU = (unsigned char*)_mm_malloc(widthUPadded*heightU_itlc*sizeof(char),128);
		InitTable(widthUPadded, heightU_itlc, nsteps, LUTU);
	}

}

//----------------------------------------------------------------------------------
VagueDenoiser::~VagueDenoiser()
{
	_mm_free(block);
	if (Wiener) _mm_free(block1);// v.0.30

	if(!vi.IsYV12())
	{
		_mm_free(buffY); // add to fix possible memory leakage bug.-  v.0.30 by Fizick
		_mm_free(buffU);
		_mm_free(buffV);
		if (useAuxClip) // v 0.34
		{
			_mm_free(auxbuffY); 
			_mm_free(auxbuffU);
			_mm_free(auxbuffV);
		}
	}
	if (itlc) _mm_free(buffer);

	delete wavelet;
	if (Wiener) delete wavelet1; // v.0.30

	if (method == 2) // v0.35
	{
		_mm_free(LUTY); 
		_mm_free(LUTU); 
	}
}

//----------------------------------------------------------------------------------
void VagueDenoiser::filterBlockYUY2(const unsigned char* src,
									unsigned char* dst, int srcPitch, int dstPitch,
									const unsigned char* aux, int auxPitch) //changed in v.0.34
{
	int i,j;    
	unsigned char* buffptr=buffY;
	unsigned char* buffptrU=buffU;
	unsigned char* buffptrV=buffV;

	unsigned char* auxbuffptr=auxbuffY;
	unsigned char* auxbuffptrU=auxbuffU;
	unsigned char* auxbuffptrV=auxbuffV;
	
	if (chromaT >= 0 && lumaT >= 0)
	{
		//Copy Source to Buffer
		for (i=0; i < height;i++)
		{
			int k=0;
			for (j = 0; j < (widthU<<2); j+=4)
			{
				buffptr[k<<1] = src[j];
				buffptrU[k] = src[j+1];
				buffptr[(k<<1)+1]= src[j+2];
				buffptrV[k] = src[j+3];
				k++;
			}
			src += srcPitch;
			buffptr += pitch;//width;  changed in 0.34.2
			buffptrU+= pitchU;//widthU; changed in 0.34.2
			buffptrV+= pitchU;//widthU; changed in 0.34.2
		}

		if (useAuxClip) // added in v.0.34
		{
		//Copy Aux to auxBuffer
			for (i=0; i < height;i++)
			{
				int k=0;
				for (j = 0; j < (widthU<<2); j+=4)
				{
					auxbuffptr[k<<1] = aux[j];
					auxbuffptrU[k] = aux[j+1];
					auxbuffptr[(k<<1)+1]= aux[j+2]; // fixed bug in 0.34.2
					auxbuffptrV[k] = aux[j+3];
					k++;
				}
				aux += auxPitch;
				auxbuffptr += pitch;//width; changed in 0.34.2
				auxbuffptrU+= pitchU;//widthU; changed in 0.34.2
				auxbuffptrV+= pitchU;//widthU; changed in 0.34.2
			}
		}

		//set threshold value to lumaT
		thresh = lumaT; // Added in v.0.30 by Fizick
		filterBlockPlane(height, width, buffY, buffY, pitch, pitch, auxbuffY, pitch);//changed in v.0.34.2

		//set threshold value to chromaT
		thresh = chromaT; // Added in v.0.30 by Fizick
		filterBlockPlane(heightU, widthU, buffU, buffU, pitchU, pitchU, auxbuffU, pitchU);//changed in v.0.34
		filterBlockPlane(heightU, widthU, buffV, buffV, pitchU, pitchU, auxbuffV, pitchU);

		buffptr=buffY;
		buffptrU=buffU;
		buffptrV=buffV;
		for (i=0; i < height;i++)
		{
			int k=0;
			for (j = 0; j < widthU; j++)
			{
				dst[k]  = buffptr[j<<1];
				dst[k+1]= buffptrU[j];
				dst[k+2]= buffptr[(j<<1)+1];
				dst[k+3]= buffptrV[j];
				k+=4;
			}
			dst += dstPitch;
			buffptr += pitch;//width;  changed in 0.34.2
			buffptrU+= pitchU;//widthU; changed in 0.34.2
			buffptrV+= pitchU;//widthU; changed in 0.34.2
		}
	}
	else if (lumaT >= 0)
	{
		//Copy Source to Buffer
		for (i=0; i < height;i++)
		{
			int k=0;
			for (j = 0; j < (widthU<<2); j+=4)
			{
				buffptr[k<<1] = src[j];
				buffptr[(k<<1)+1]= src[j+2];
				k++;
			}
			src += srcPitch;
			buffptr += pitch;// width; changed in 0.34.2
		}

		if (useAuxClip)
		{
		//Copy Aux to auxBuffer
			for (i=0; i < height;i++)
			{
				int k=0;
				for (j = 0; j < (widthU<<2); j+=4)
				{
					auxbuffptr[k<<1] = aux[j];
					auxbuffptr[(k<<1)+1]= aux[j+2]; // fixed bug in v0.35.1
					k++;
				}
				aux += auxPitch;
				auxbuffptr += pitch;// width; changed in 0.34.2
			}
		}


		//set threshold value to lumaT
		thresh = lumaT; // Added in v.0.30 

		filterBlockPlane(height, width, buffY, buffY, pitch, pitch, auxbuffY, pitch);

		buffptr=buffY;
		for (i=0; i < height;i++)
		{
			int k=0;
			for (j = 0; j < widthU; j++)
			{
				dst[k]  = buffptr[j<<1];
				dst[k+2]= buffptr[(j<<1)+1];
				k+=4;
			}
			dst += dstPitch;
			buffptr += pitch;// width; changed in 0.34.2
		}
	}
	else if (chromaT >= 0)
	{
		//Copy Source to Buffer
		for (i=0; i < height;i++)
		{
			int k=0;
			for (j = 0; j < (widthU<<2); j+=4)
			{
				buffptrU[k] = src[j+1];
				buffptrV[k] = src[j+3];
				k++;
			}
			src += srcPitch;
			buffptrU+= pitchU;// widthU; changed in 0.34.2
			buffptrV+= pitchU;// widthU; changed in 0.34.2
		}

		if (useAuxClip)
		{
		//Copy Aux to auxBuffer
			for (i=0; i < height;i++)
			{
				int k=0;
				for (j = 0; j < (widthU<<2); j+=4)
				{
					auxbuffptrU[k] = aux[j+1];
					auxbuffptrV[k] = aux[j+3];
					k++;
				}
				aux += auxPitch;
				auxbuffptrU+= pitchU;// widthU; changed in 0.34.2
				auxbuffptrV+= pitchU;// widthU; changed in 0.34.2
			}
		}


		//set threshold value to chromaT
		thresh = chromaT; // Added in v.0.30 by Fizick
		filterBlockPlane(heightU, widthU, buffU, buffU, pitchU, pitchU, auxbuffU, pitchU); //changed in 0.34.2
		filterBlockPlane(heightU, widthU, buffV, buffV, pitchU, pitchU, auxbuffV, pitchU); //changed in 0.34.2

		buffptrU=buffU;
		buffptrV=buffV;
		for (i=0; i < height;i++)
		{
			int k=0;
			for (j = 0; j < widthU; j++)
			{
				dst[k+1]= buffptrU[j];
				dst[k+3]= buffptrV[j];
				k+=4;
			}
			dst += dstPitch;
			buffptrU+= pitchU;// widthU;changed in 0.34.2
			buffptrV+= pitchU;// widthU;changed in 0.34.2
		}
	}
}
//----------------------------------------------------------------------------------
void VagueDenoiser::filterBlockRGB(const unsigned char* src, unsigned char* dst,
								   int srcPitch, int dstPitch, char numb,
								   const unsigned char* aux, int auxPitch)//changed in v.0.34
{
	int i,j;    
	unsigned char* buffptr=buffY;
	unsigned char* buffptrU=buffU;
	unsigned char* buffptrV=buffV;
	unsigned char* auxbuffptr=auxbuffY;
	unsigned char* auxbuffptrU=auxbuffU;
	unsigned char* auxbuffptrV=auxbuffV;
	
	//Copy Source to Buffer
	for (i=0; i < height;i++)
	{
		int k=0;
		for (j = 0; j < width; j++)
		{
			buffptr[j] = src[k];
			buffptrU[j] = src[k+1];
			buffptrV[j] = src[k+2];
			k+=numb;
		}
		src += srcPitch;
		buffptr += pitch;//width; 0.34.2
		buffptrU += pitchU;//width;
		buffptrV += pitchU;//width;
	}

	if (useAuxClip) //added in v.0.34
	{
	//Copy Aux to auxBuffer
		for (i=0; i < height;i++)
		{
			int k=0;
			for (j = 0; j < width; j++)
			{
				auxbuffptr[j] = aux[k];
				auxbuffptrU[j] = aux[k+1];
				auxbuffptrV[j] = aux[k+2];
				k+=numb;
			}
			aux += auxPitch;
			auxbuffptr += pitch;//width;
			auxbuffptrU += pitchU;//width;
			auxbuffptrV += pitchU;//width;
		}
	}

	filterBlockPlane(height, width, buffY, buffY, pitch, pitch, auxbuffY, pitch);//changed in v.0.34.2
	filterBlockPlane(heightU, widthU, buffU, buffU, pitchU, pitchU, auxbuffU, pitchU);
	filterBlockPlane(heightU, widthU, buffV, buffV, pitchU, pitchU, auxbuffV, pitchU);

	buffptr=buffY;
	buffptrU=buffU;
	buffptrV=buffV;
	for (i=0; i < height;i++)
	{
		int k=0;
		for (j = 0; j < width; j++)
		{
			dst[k]  = buffptr[j];
			dst[k+1]= buffptrU[j];
			dst[k+2]= buffptrV[j];
			k+=numb;
		}
		dst += dstPitch;
		buffptr += pitch;//width;
		buffptrU+= pitchU;//widthU;
		buffptrV+= pitchU;//widthU;
	}
}

//---------------------------------------------------------------------------------------
void VagueDenoiser::filterBlockPlane(int h, int w0,  // changed to w0 in v.0.33
									const unsigned char* src,
									unsigned char* dst, int srcPitch, int dstPitch,
									const unsigned char* aux, int auxPitch) // added in v.0.34
{

		 
	// add pad to width to prevent cache conflicts for width=mod64 - added in v.0.33
	int w = (w0%64) ? w0 : w0 + WIDTHPAD; 
	// w is now padded. it is width (or pitch) of float block used

	unsigned char* LUT; // v0.35
	if (w0 == width) 
		LUT = LUTY;
	else // w0 == widthU 
		LUT = LUTU;

	if (!itlc) // progressive
	{
		// Most usual case, always this one for YUY2 and RGBx 
		//if (w0 != srcPitch) byte2float(h, w, src, srcPitch, block);
		//else byte2float1D(w*h, src, block);
		byte2float(h, w0, src, srcPitch, block, w); // changed in v.0.33

		if(w>w0) mirror_pad(block, h, w); // mirror rigthmost padded colunms - added in v.0.33, bug fixed (condition added) in v.0.33.4

		//Apply filtering
		
		if (!Wiener)
		{ // standard one-pass mode
			wavelet->transform2d(block,w,h,nsteps);	
			if (thresh == 0.0f) AutoThreshold(block, w, h, nsteps, wratio, &thresh); // added in v.0.30 by Fizick

			if (debug)
			{
				wsprintf(debugbuf,"VagueDenoiser: w=%d h=%d thresh=%e\n", w, h, thresh);
				OutputDebugString(debugbuf);
			}
	
			if (method != -1 && method !=2 ) filter(block, w*h, thresh, w, h, nsteps, percent); // this line was occasionally commented in 0.29 source
			else if ( method ==2 ) MultiScaleThresholding(block, w*h, thresh, w, h, nsteps, wratio, percent, LUT);//v0.35
			if (!debug)	wavelet->invert2d(block,w,h,nsteps);
		}
		else
		{   // WienerChop two-pass mode - Added in v.0.30 by Fizick
			if (useAuxClip) // added in v.0.34
			{ // use aux clip frame as first pass clip
				byte2float(h, w0, aux, auxPitch, block1, w); // aux frame bytes to float block1
				if(w>w0) mirror_pad(block1, h, w); // mirror rigthmost padded colunms - added in v.0.33, bug fixed (condition added) in v.0.33.4
			}
			else 
			{   // standard pre v.0.34 mode without aux clip: copy block (i.e. source) to block1
				VDEnv->BitBlt( (BYTE*)block1, w*sizeof(float), (BYTE*) block, w*sizeof(float), w*sizeof(float), h); 
			}
			// first pass in first wavelet basis
			wavelet1->transform2d(block1,w,h,nsteps); // transform source  in first wavelet basis
			if (thresh == 0.0f) AutoThreshold(block1, w, h, nsteps, wratio, &thresh);
			if (debug)
			{
				wsprintf(debugbuf,"VagueDenoiser: w=%d h=%d thresh=%e\n", w, h, thresh);
				OutputDebugString(debugbuf);
			}
			if (method != -1 && method !=2 ) filter(block1, w*h, thresh, w, h, nsteps, percent); // thresholding
			else if ( method ==2 ) MultiScaleThresholding(block1, w*h, thresh, w, h, nsteps, wratio, percent, LUT);//v0.35
			wavelet1->invert2d(block1,w,h,nsteps); // // get first-pass denoised image. It will be used as estimation 
			// second pass - transform in second wavelet basis
			wavelet->transform2d(block1,w,h,nsteps);	// transform estimation
			wavelet->transform2d(block,w,h,nsteps);	// transform source image
			if (method != -1) WienerChop(block1, block, w*h, thresh*wratio);// Wiener filter of source
			wavelet->invert2d(block,w,h,nsteps); // get denoised image
		}

		// Float to byte 
		//if (w != dstPitch) float2byte(h, w, dst, dstPitch, block);
		//else float2byte1D(h*w, dst, block);
		float2byte(h, w0, dst, dstPitch, block,w); // changed in v.0.33
//		if(percent != 100) partial_denoise( dst, dstPitch, src, srcPitch, w0, h, percent);
	}
	else // interlaced
	{
		// source fields to buffer
		VDEnv->BitBlt(buffer, w0,  src, srcPitch<<1, w0, h>>1); 
		VDEnv->BitBlt(buffer+w0*(h>>1), w0, src+srcPitch, srcPitch<<1, w0, h>>1); 
		//Bytes to float block
		byte2float(h, w0, buffer, w0, block,w); // changed in v.0.33
		if(w>w0) mirror_pad(block, h, w); // mirror rigthmost padded  colunms - added (was omitted) in v.0.33.4

		if (!Wiener)
		{ // standard one-pass mode
			   // First field 
			wavelet->transform2d(block,w,h>>1,nsteps);
			if (thresh == 0)  AutoThreshold(block, w, h>>1, nsteps, wratio, &thresh);
			if (debug)
			{
				wsprintf(debugbuf,"VagueDenoiser: w=%d h=%d thresh=%e\n", w, h, thresh);
				OutputDebugString(debugbuf);
			}
			if (method != -1 && method != 2 ) filter(block, w*(h>>1), thresh, w, h>>1, nsteps, percent);
			else if ( method ==2 ) MultiScaleThresholding(block, w*(h>>1), thresh, w, h>>1, nsteps, wratio, percent, LUT); //v0.35
			if (!debug)	wavelet->invert2d(block,w,h>>1,nsteps);

			  // Second field
			wavelet->transform2d(block+w*(h>>1),w,h>>1,nsteps);
			if (method != -1 && method !=2 ) filter(block+w*(h>>1), w*(h>>1), thresh, w, h>>1, nsteps, percent);
			else if ( method ==2 ) MultiScaleThresholding(block+w*(h>>1), w*(h>>1), thresh, w, h>>1, nsteps, wratio, percent, LUT); //v0.35
			if (!debug)	wavelet->invert2d(block+w*(h>>1),w,h>>1,nsteps);
		}
		else
		{
			// WienerChop two-pass mode - Added in v.0.30 by Fizick
			if (useAuxClip) // added in v.0.34
			{ // use aux clip frame as first pass clip
				// aux to buffer
				VDEnv->BitBlt(buffer, w0,  aux, auxPitch<<1, w0, h>>1); 
				VDEnv->BitBlt(buffer+w0*(h>>1), w0, aux+auxPitch, auxPitch<<1, w0, h>>1); 
				//Byte to float block1
				byte2float(h, w0, buffer, w0, block1,w); 
				if(w>w0) mirror_pad(block1, h, w); // mirror rigthmost padded  colunms - added (was omitted) in v.0.33.4
			}
			else 
			{   // standard pre v.0.34 mode without aux clip: 
				//copy block (i.e. source) to block1
				VDEnv->BitBlt( (BYTE*)block1, w*sizeof(float), (BYTE*) block, w*sizeof(float), w*sizeof(float), h); 
			}
			// First field 
			// first pass in first wavelet basis
			wavelet1->transform2d(block1,w,h>>1,nsteps); // transform source  in first wavelet basis
			if (thresh == 0) AutoThreshold(block1, w, h>>1, nsteps, wratio, &thresh); // bug fixed in v.0.33.5
			if (debug)
			{
				wsprintf(debugbuf,"VagueDenoiser: w=%d h=%d thresh=%e\n", w, h, thresh);
				OutputDebugString(debugbuf);
			}
			if (method != -1 && method != 2 )filter(block1, w*(h>>1), thresh, w, h>>1, nsteps, percent); // thresholding
			else if ( method ==2 ) MultiScaleThresholding(block1, w*(h>>1), thresh, w, h>>1, nsteps, wratio, percent, LUT); //v0.35
			wavelet1->invert2d(block1,w,h>>1,nsteps); // // get first-pass denoised image. It will be used as estimation 
			// second pass - transform in second wavelet basis
			wavelet->transform2d(block1,w,h>>1,nsteps);	// transform estimation
			wavelet->transform2d(block,w,h>>1,nsteps);	// transform source image
			if (method != -1) WienerChop(block1, block, w*(h>>1), thresh*wratio);// Wiener filter of source
			wavelet->invert2d(block,w,(h>>1),nsteps); // get denoised image

			// Second field 
			// first pass in first wavelet basis
			wavelet1->transform2d(block1+w*(h>>1),w,h>>1,nsteps); // transform source  in first wavelet basis
			if (method != -1 && method != 2 )filter(block1+w*(h>>1), w*(h>>1), thresh, w, h>>1, nsteps, percent); // thresholding
			else if ( method ==2 ) MultiScaleThresholding(block1+w*(h>>1), w*(h>>1), thresh, w, h>>1, nsteps, wratio, percent, LUT); //v0.35
			wavelet1->invert2d(block1+w*(h>>1),w,h>>1,nsteps); // // get first-pass denoised image. It will be used as estimation 
			// second pass - transform in second wavelet basis
			wavelet->transform2d(block1+w*(h>>1),w,h>>1,nsteps);	// transform estimation
			wavelet->transform2d(block+w*(h>>1),w,h>>1,nsteps);	// transform source image
			if (method != -1) WienerChop(block1+w*(h>>1), block+w*(h>>1), w*(h>>1), thresh*wratio);// Wiener filter of source
			wavelet->invert2d(block+w*(h>>1),w,(h>>1),nsteps); // get denoised image

		}

		//Float to byte
		//if (w0 != dstPitch) float2byte(h, w, buffer, w, block); // bug fixed in v 0.31
		//else float2byte1D(h*w, buffer, block);                 // bug fixed in v 0.31
		float2byte(h, w0, buffer, w0, block, w); // changed in v.0.33

		VDEnv->BitBlt(dst, dstPitch<<1, buffer, w0, w0, h>>1); 
		VDEnv->BitBlt(dst+dstPitch, dstPitch<<1, buffer+w0*(h>>1), w0, w0, h>>1); 

//		if(percent != 100) // partial denoising
		{
//			partial_denoise( dst, dstPitch<<1, src, srcPitch<<1, w0, h>>1, percent);
//			partial_denoise( dst+dstPitch, dstPitch<<1, src+srcPitch, srcPitch<<1, w0, h>>1, percent);
		}
	}
}
//----------------------------------------------------------------------------------------
PVideoFrame __stdcall VagueDenoiser::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame src = child->GetFrame(n, env);

	PVideoFrame aux; // v.0.34
	if (useAuxClip)
		aux = auxClip->GetFrame(n,env);
	else
		aux = src;

	if (debug)
	{
		wsprintf(debugbuf,"VagueDenoiser: frame=%d\n", n);
		OutputDebugString(debugbuf);
	}
	
	if(vi.IsYV12())
	{
		PVideoFrame dst = env->NewVideoFrame(vi); // v.0.32 - it is more clear. makewritable simply make a copy.
		if(lumaT >= 0)
		{
			//set threshold value to lumaT
			thresh = lumaT; // Added in v.0.30 by Fizick
			//Filter Y Plane
			filterBlockPlane(height, width, src->GetReadPtr(PLANAR_Y),
							dst->GetWritePtr(PLANAR_Y), src->GetPitch(PLANAR_Y),
							dst->GetPitch(PLANAR_Y),
							aux->GetReadPtr(PLANAR_Y), aux->GetPitch(PLANAR_Y)); // v.0.34						
		}
		else
		{
			env->BitBlt(dst->GetWritePtr(PLANAR_Y), dst->GetPitch(PLANAR_Y), 
				src->GetReadPtr(PLANAR_Y), src->GetPitch(PLANAR_Y), //  fixed in 0.33.3
				width, height);
		}

		if(chromaT >= 0)
		{
			//set threshold value to chromaT
			thresh = chromaT; // Added in v.0.30 by Fizick
			//Filter U Plane
			filterBlockPlane(heightU, widthU, src->GetReadPtr(PLANAR_U),
							dst->GetWritePtr(PLANAR_U), src->GetPitch(PLANAR_U),
							dst->GetPitch(PLANAR_U),
							aux->GetReadPtr(PLANAR_U), aux->GetPitch(PLANAR_U));						
			
			//Filter V Plane
			filterBlockPlane(heightU, widthU, src->GetReadPtr(PLANAR_V),
							dst->GetWritePtr(PLANAR_V), src->GetPitch(PLANAR_V),
							dst->GetPitch(PLANAR_V),
							aux->GetReadPtr(PLANAR_V), aux->GetPitch(PLANAR_V));						
		}
		else
		{
			env->BitBlt(dst->GetWritePtr(PLANAR_U), dst->GetPitch(PLANAR_U), 
				src->GetReadPtr(PLANAR_U), src->GetPitch(PLANAR_U), //  fixed in 0.33.3
				widthU, heightU);
			env->BitBlt(dst->GetWritePtr(PLANAR_V), dst->GetPitch(PLANAR_V), 
				src->GetReadPtr(PLANAR_V), src->GetPitch(PLANAR_V), //  fixed in 0.33.3
				widthU, heightU);
		}
		return dst;
	}
	else if(vi.IsYUY2())
	{
		if (lumaT >= 0 && chromaT >=0 )
		{
			PVideoFrame dst = env->NewVideoFrame(vi); // v.0.32 - it is more clear. makewritable simply make a copy.
		
			filterBlockYUY2(src->GetReadPtr(), dst->GetWritePtr(),
							src->GetPitch(), dst->GetPitch(),
							aux->GetReadPtr(), aux->GetPitch()); //  fixed in 0.33.3
			return dst;
		}
		else if (lumaT >= 0 || chromaT >=0 )
		{
			PVideoFrame dst = env->NewVideoFrame(vi); // v.0.32 - it is more clear. makewritable simply make a copy.
			env->BitBlt(dst->GetWritePtr(), dst->GetPitch(), 
				src->GetReadPtr(), src->GetPitch(), 
				src->GetRowSize(), src->GetHeight());// it is some over, but fast enough - bug fixed in v.0.33.3
			filterBlockYUY2(src->GetReadPtr(), dst->GetWritePtr(),//  fixed in 0.33.3
							src->GetPitch(), dst->GetPitch(),
							aux->GetReadPtr(), aux->GetPitch());
			return dst;
		}
		else
			return src;
	}
	else // RGB
	{

		if (lumaT >=0) 
		{
//			env->MakeWritable(&src);
			PVideoFrame dst = env->NewVideoFrame(vi); // v.0.32 - it is more fast. makewritable simply make a copy.
			//set threshold value to lumaT
			thresh = lumaT; // Added in v.0.30 by Fizick
			filterBlockRGB(src->GetReadPtr(), dst->GetWritePtr(),
						src->GetPitch(), dst->GetPitch(), vi.BitsPerPixel()>>3,
						aux->GetReadPtr(), aux->GetPitch());
			return dst;
		}
		else
			return src;
	}
} 

/*************************      Externs      ***************************/
									 
AVSValue __cdecl Create_VagueDenoiser(AVSValue args, void* user_data,
									  IScriptEnvironment* env)
{
	long CPUFlags = env->GetCPUFlags(); // v.0.33.6
	
	try // added for exception processing in  v.0.33.6
	{

	if ((CPUFlags & CPUF_3DNOW_EXT) &&	(CPUFlags & CPUF_SSE))
		return new VagueDenoiser3DNowExt(args[0].AsClip(),
										 args[1].AsFloat(0.0),	args[2].AsInt(3), // v.0.332
										 args[3].AsInt(4), args[4].AsFloat(-1.0f),// v.0.332
										 args[5].AsBool(false), args[6].AsBool(false),
										args[7].AsInt(1),  args[8].AsBool(false),// v.0.30 by Fizick
										args[9].AsFloat(0.5f), // v.0.31 by Fizick
										args[10].AsFloat(85),// v.0.33 by Fizick
										args[11].AsClip(),// v.0.34 by Fizick
										 env);
	else if (CPUFlags & CPUF_SSE)
		return new VagueDenoiserSSE(args[0].AsClip(),
									  args[1].AsFloat(0.0),	args[2].AsInt(3),
									  args[3].AsInt(4), args[4].AsFloat(-1.0f),
									  args[5].AsBool(false), args[6].AsBool(false),
									  args[7].AsInt(1),  args[8].AsBool(false),// v.0.30 by Fizick
									  args[9].AsFloat(0.5f), // v.0.31 by Fizick
									  args[10].AsFloat(85),// v.0.33 by Fizick
										args[11].AsClip(),// v.0.34 by Fizick
									  env);
	else if (CPUFlags & CPUF_3DNOW)
		return new VagueDenoiser3DNow(args[0].AsClip(),
									  args[1].AsFloat(0.0),	args[2].AsInt(3),
									  args[3].AsInt(4), args[4].AsFloat(-1.0f),
									  args[5].AsBool(false), args[6].AsBool(false),
									  args[7].AsInt(1),  args[8].AsBool(false),// v.0.30 by Fizick
									  args[9].AsFloat(0.5f), // v.0.31 by Fizick
									  args[10].AsFloat(85),// v.0.33 by Fizick
										args[11].AsClip(),// v.0.34 by Fizick
									  env);
	else
		return new VagueDenoiserC(args[0].AsClip(),
									  args[1].AsFloat(0.0),	args[2].AsInt(3),
									  args[3].AsInt(4), args[4].AsFloat(-1.0f),
									  args[5].AsBool(false),  args[6].AsBool(false),
									  args[7].AsInt(1),  args[8].AsBool(false),// v.0.30 by Fizick
									  args[9].AsFloat(0.5f), // v.0.31 by Fizick
									  args[10].AsFloat(85),// v.0.33 by Fizick
										args[11].AsClip(),// v.0.34 by Fizick
									  env);
	}
	catch (...)
	{
		throw;
	}
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) {
	env->AddFunction("VagueDenoiser", "c[threshold]f[method]i[nsteps]i[chromaT]f[debug]b[interlaced]b[wavelet]i[Wiener]b[wratio]f[percent]f[auxClip]c", Create_VagueDenoiser, 0);
	return "`VagueDenoiser' VagueDenoiser plugin";
}



/**********************       C Stuff       ****************************/
_inline unsigned char realToChar (float x) { 
	return (x>255.0f?255:(x<0.0f?0:(unsigned char)(x+0.5f)));}
_inline float sgn (float x) {
	return (x<0)?-1.0f:((x>0)?1.0f:0.0f);}

/**** Some C version of thresholding functions ****/
void ht_c(float* block, const unsigned int b, const float threshold,
		  const unsigned int width, const unsigned int height, const unsigned int nsteps, float percent)
{
	// hard thresholding
	float frac = 1 - percent*0.01f; // partial denoise added by Fizick in v0.35 

	for(unsigned int i=0;i<b;i++)
		if (fabsf(block[i]) <= threshold) block[i] *= frac;
}

void st_c(float* block, const unsigned int b, const float threshold,
		  const unsigned int width, const unsigned int height, const unsigned int nsteps, float percent)
{
	// soft thresholding (shrinking)
	float frac = 1 - percent*0.01f; // partial denoise added by Fizick in v0.35 
	float shift = threshold*0.01f*percent;

	unsigned int h=height;
	unsigned int w=width;
	for (unsigned int l=0; l<nsteps; l++)
	{
		h = (h+1)>>1;
		w = (w+1)>>1;
	}

	for (unsigned int j=0; j<height; j++)
	{
		unsigned int i0 = (j<h) ? w : 0; 
		// skip lowest level to preserve mean intensity - v0.35
		for(unsigned int i=i0; i<width; i++)
		{
			float temp=fabsf(block[i]);
			if (temp <= threshold) block[i] *= frac;
			else block[i]= sgn(block[i])*(temp-shift);
		}
		block += width;
	}
}

void qt_c(float* block, const unsigned int b, const float threshold,
		  const unsigned int width, const unsigned int height, const unsigned int nsteps, float percent)
{
	// Qian (garrote) thresholding (shrinking)
	float tr2 = threshold*threshold;
	float percent01 = percent*0.01f; // partial denoise added by Fizick in v0.35  
	tr2 *= percent01;
	float frac = 1 - percent01;

	for(unsigned int i=0;i<b;i++)
	{
		float temp=fabsf(block[i]);
		if(temp<=threshold) block[i] *= frac;
		else
		{
			float tp2 = temp*temp;
			block[i]*=(tp2-tr2)/tp2;
		}
	}
}

void smt_c(float* block, const unsigned int b, const float threshold,
		  const unsigned int width, const unsigned int height, const unsigned int nsteps, float percent)
{
	// uniformly smooth shrinking function by Fizick - added in v0.35
	float sfactor = 2.0f; // empirical factor for similarity with other thtresholding functions
	float tr2 = threshold*threshold*sfactor;
	float frac = 1 - percent*0.01f; // partial denoise added by Fizick in v0.35  
	float tr2frac = tr2*frac;
	for(unsigned int i=0;i<b;i++)
	{
		float temp = block[i];
		float tp2 = temp*temp;
		block[i] = temp*(tp2+tr2frac)/(tp2 + tr2);
	}
}
//--------------------------------------------------------------------------
void b2f_c(const unsigned int h, const unsigned int w,
		   const unsigned char* src, const unsigned int srcPitch,
		   float *blockptr,
		   const unsigned int blockPitch) //  blockpitch added in v.0.33
{
	//Copy Source to Buffer
	for (unsigned int i=0; i < h;i++)
	{
		for (unsigned int j = 0; j < w; j+=2)
		{
			blockptr[j] =  src[j];
			blockptr[j+1] = src[j+1];
		}
		src += srcPitch;
//		blockptr += w;
		blockptr += blockPitch;
	}
}

void f2b_c(const unsigned int h, const unsigned int w,
		   unsigned char *dst, const unsigned int dstPitch,
		   float *blockptr,
		   const unsigned int blockPitch) //  blockpitch added in v.0.33
{
	//Copy Buffer to Destination
	for (unsigned int i=0; i < h;i++)
	{
		for (unsigned int j = 0; j < w; j+=2)
		{
			dst[j] = realToChar(blockptr[j]);
			dst[j+1] = realToChar(blockptr[j+1]);
		}
		dst += dstPitch;
//		blockptr += w;
		blockptr += blockPitch;
	}
}

void b2f1D_c(const unsigned int w, const unsigned char* src, float *blockptr)
{
	//Copy Source to Buffer
	for (unsigned int i=0; i < w;i+=2)
	{
		blockptr[i] =  src[i];
		blockptr[i+1] =  src[i+1];
	}
}

void f2b1D_c(const unsigned int w, unsigned char *dst, float *blockptr)
{
	//Copy Buffer to Destination
	for (unsigned int i=0; i < w;i+=2)
	{
		dst[i] = realToChar(blockptr[i]);
		dst[i+1] = realToChar(blockptr[i+1]);
	}
}

void wiener_chop_c(const float* block1, float *block, const unsigned int length, float sigmaNoise)
{
	// added in v.0.30, modified in v.0.31, renamed in Kurosu's style in 0.32
	// Wiener filter of block with variation estimated from block1 
//	float sigmaNoise = (threshold/3); // noise std dev. Scale is taken similar to one-pass method 
	float varNoise = sigmaNoise * sigmaNoise; // noise variation. 
	for (unsigned int i=0; i<length; i++)
	{
		float varEstimation = block1[i]*block1[i];// estimated data used as variation
		block[i] *= varEstimation/(varNoise + varEstimation); // Wiener filter
	}
}
//#include "stdio.h" // for debug
//#include "windows.h"

void auto_th_c(const float*block, int w, int h, int nsteps, float ratio, float * thresholdEstimated)
{
	// added in v.0.30 by Fizick, modified in v.0.31, renamed in Kurosu's style in 0.32
	// fixed in v0.34.3 (was not finest level)
	// automatic (adaptive) estimation of threshold
	// by Mean absolute deviation of finest level of wavelet transformed block
	// Probably Median absolute deviation will be better? (with qsort, may be in next version)
	 int length4 = w*h/4;
	 int w2 = w/2;
	 int h2 = h/2;
	float MAD = 0;
	block += w*h/2 + w2; // start position
	for ( int j=h2; j<h; j++)
	{
		for ( int i=0; i<w2; i++) // finest level
		{
			MAD += fabsf(block[i]); // sum of mean absolute deviations
		}
		block += w;
	}
	MAD /= length4; // mean absolute deviation

	float sigmaNoise = MAD/0.6745f; // well known relation
	*thresholdEstimated = sigmaNoise/ratio; // was =2 - changed in v.0.33

//	char debugbuf[80];
//	sprintf(debugbuf,"Vague: w=%d h=%d th=%e\n", w, h, *thresholdEstimated);
//	OutputDebugString(debugbuf);
}

void mirror_pad(float *blockptr, const unsigned int h, const unsigned int w) // added in v.0.33

{ // mirror rigthmost (above w0) widthPad columns to prevent oscillations near edge
	// w is full width  with widthpad (pitch)
	// widthpad is now  here to unroll loop 
	// MMX version would be simple but it is not main time loop
	for (unsigned int i=0; i < h;i++)
	{
		blockptr[w-WIDTHPAD] =  blockptr[w-WIDTHPAD-1]; // start
			// no middle now
		blockptr[w-3] =  blockptr[w-WIDTHPAD-2]; // added in v.0.33.1 (pad = 4)
		blockptr[w-2] =  blockptr[w-WIDTHPAD-3]; // added in v.0.33.1 (pad = 4)
		blockptr[w-1] =  blockptr[w-WIDTHPAD-WIDTHPAD]; //end
		blockptr += w;
	}
}
/*
void partial_denoise( unsigned char* dst, int dstPitch, const unsigned char* src, int srcPitch,
			  unsigned int w, unsigned int h, float percentDst) // removed (replaced) in v.0.35
{
	// partial_denoise destination (denoised) plane with source by blending
	int fractSrc = ((100-(int)percentDst)<<8)/100;
	
	for (unsigned int i=0; i < h; i++)
	{
//		for (unsigned int j = 0; j < w; j++) 
		for (int j = w-1; j >=0; j--) // reverse order give faster code
		{
			dst[j] += ( (src[j]-dst[j])*fractSrc)>>8;
		}
		dst += dstPitch;
		src += srcPitch;
	}

} */
//----------------------------------------------------------------------------------------------
void InitTable(int width, int height, int nsteps, unsigned char* LUT) 
// init subband indexes - added in v0.35 (restored from v.0.28)
{
	int x;
	int xmax = width;
//	int xmin = width>>1;
	int xmin = (width+1)>>1; // it is more correct here and below (see transform function) - v0.35
	int y;
	int ymax = height;
//	int	ymin = height>>1;
	int	ymin = (height+1)>>1;
	int i;
	unsigned char *LUTp;
	
//	LUT = (unsigned char*)malloc(width*height*sizeof(char)); // moved to main program
//	maxs = (float *)malloc((1+3*(nsteps+1))*sizeof(float));

	i = nsteps;//nsteps+1;
	while (i>0)
	{
		
		LUTp = LUT;
		for (y=0; y<ymin; y++)
		{
			for (x=xmin; x<xmax; x++)
				LUTp[x] = 3*i-2;
			LUTp += width;
		}
		
		LUTp = LUT + ymin*width;
		for (y=ymin; y<ymax; y++)
		{
			for (x=0; x<xmin; x++)
				LUTp[x] = 3*i-1;
			
			for (x=xmin; x<xmax; x++)
				LUTp[x] = 3*i;
			
			LUTp += width;
		}
		
		i--;
		xmax = xmin;
//		xmin >>= 1;
		xmin = (xmin+1)>> 1;
		ymax = ymin;
//		ymin >>= 1;
		ymin = (ymin+1)>> 1;
	}

	//Level = 0
	LUTp = LUT;
	for (y=0; y<ymax; y++)
	{
		for (x=0; x<xmax; x++)
			LUTp[x] = 0;
		LUTp += width;
	}

}

void MultiScaleThresholding(float *block, int length, float thresh, int width, int height, int nsteps, float ratio, float percent, unsigned char *LUT)
{
// A litle modified NormalShrink method added in v0.35 (partially based on v.0.28 code) 
// Image Denoising using Wavelet Thresholding. 
// Lakhwinder Kaur, Savita Gupta, R.C. Chauhan.
// Third Conference on Computer Vision, Graphics and Image Processing, India Dec 16-18, 2002.
// http://www.ee.iitb.ac.inq/~icvgip/PAPERS/202.pdf

// see also related BayesShrink method 
// S. Grace Chang, Bin Yu, and Martin Vetterli. 
// Adaptive Wavelet Thresholding for Image Denoising and Compression.
// IEEE TRANSACTIONS ON IMAGE PROCESSING, VOL. 9, NO. 9, SEPTEMBER 2000, p.1532-1546.
	int i;
	float frac = 1 - percent*0.01f; // partial denoise added by Fizick in v0.35 
	float percent01 = percent*0.01f;
	float sigma = thresh*ratio; // noise value
	float sigma2 = sigma*sigma; // noise variance

	float * esp = (float *)calloc((1+3*(nsteps+1)),sizeof(float));
	unsigned int *cnt = (unsigned int *)calloc((1+3*(nsteps+1)),sizeof(unsigned int));

	// calloc must init arrays by zeros, but we can make it manully to be sure
	for (i=0; i<1+(nsteps)*3; i++)
	{
		esp[i] =1e-6f; // to prevent possible divide below by null for pure black
		cnt[i] =0;
	}

	// I use LUT here (as its code by Kurosu was ready), 
	// but explicit subband boundaries could be faster and SSE-3DNow opmimizabale.

	//Compute statistics for every subband
	for (i=0; i<length; i++)
	{
		int index = LUT[i];
		esp[index] += block[i]*block[i];
		cnt[index]++; //Could be harcoded
	}
	
	//Get thresholds for most relevant data
	  //First, deviation calculus
	for (i=1; i<1+(nsteps)*3; i++)
	{
		esp[i] /= cnt[i]; // estimate of variance of observations
		float subbandlength = sqrtf(1.0f*cnt[i]); // mean length of subband 
		float beta = sqrtf(max(logf(subbandlength),0)); //scale parameter
		// note: (subbandlenght/nsteps) was really used in the paper, but it may be too small
		esp[i] = beta*(sigma2 / sqrtf(esp[i]));	// NormalShrink subband threshold
	}

	esp[0] = 0;// do not shrink the lowest level

	// soft thresholding
	for (i=0; i<length; i++)
	{
		float thri = esp[LUT[i]];
		float shift = thri*percent01; // partial
		float babs = fabsf(block[i]);
		if (babs <= thri) 
			block[i] *= frac; 
		else
			block[i] = sgn(block[i])*(babs-shift);//soft
	}

	free(esp);
	free(cnt);
}
